r"""
    Utilities to load and build the multibase table.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

import importlib.resources as importlib_resources
import json
from typing import Dict, Iterable, Tuple, TYPE_CHECKING

from .config import _enabled_multibases

if TYPE_CHECKING:
    from multiformats.multibase import Multibase

def build_multibase_tables(bases: Iterable[Multibase]) -> Tuple[Dict[str, Multibase], Dict[str, Multibase]]:
    """
        Creates code->encoding and name->encoding mappings from a finite iterable of encodings, returning the mappings.

        Example usage:

        >>> code_table, name_table = build_multicodec_tables(bases)

        :raises ValueError: if the same encoding code or name is encountered multiple times
    """
    # pylint: disable = import-outside-toplevel
    from multiformats.multibase import Multibase
    from multiformats.multibase.err import MultibaseValueError
    # validate(multicodecs, Iterable[Multicodec]) # TODO: not yet properly supported by typing-validation
    code_table: Dict[str, Multibase] = {}
    name_table: Dict[str, Multibase] = {}
    for e in bases:
        if e.code in code_table:
            raise MultibaseValueError(f"Multicodec name {e.name} appears multiple times in table.")
        code_table[e.code] = e
        if e.name in name_table:
            raise MultibaseValueError(f"Multicodec name {e.name} appears multiple times in table.")
        name_table[e.name] = e
    return code_table, name_table

def load_multibase_table() -> Tuple[Dict[str, Multibase], Dict[str, Multibase]]:
    """
        Returns code->encoding and name->encoding mappings created (via :func:`build_multibase_tables`) from the local copy of `multibase-table.json`.
        If a subset of multibases has been enabled, only those multibases are loaded.

        Example usage:

        >>> code_table, name_table = load_multibase_table()

    """
    # pylint: disable = import-outside-toplevel
    from multiformats.multibase import Multibase
    with importlib_resources.open_text("multiformats_config", "multibase-table.json", encoding="utf8") as _table_f:
        table_json = json.load(_table_f)
        multibases = (Multibase(**row) for row in table_json)
        if _enabled_multibases is not None:
            multibases = (m for m in multibases if m.name in _enabled_multibases or m.code in _enabled_multibases)
        code_table, name_table = build_multibase_tables(multibases)
    return code_table, name_table



"""
    Configuration for the `multiformats <https://github.com/hashberg-io/multiformats>`_ package.
"""

import re
from typing import Iterable, Optional, Union

_minimal_multicodecs = frozenset([
    0x00, # identity
    0x01, # cidv1
    0x02, # cidv2
    0x12, # sha2-256
    0x14, # sha3-512
    0x16, # sha3-256
    0x70, # dag-pb
    0x71, # dag-cbor
    0x72, # libp2p-key
])

_minimal_multibases = frozenset([
    "identity",
    "base16",
    "base32",
    "base58btc",
])

_enabled_multicodecs: Optional[set[Union[int, str]]] = None
_enabled_multibases: Optional[set[str]] = None

def enable(*,
           codecs: Optional[Iterable[Union[int, str]]] = None,
           bases: Optional[Iterable[str]] = None,
          ) -> None:
    r"""
        Mantually enable a sub-set of multicodecs and/or multibases:

        - a multicodec can be specified by name (passing a :obj:`str`) or by code (passing an :obj:`int`).
        - a multibase can be specified by name (passing a :obj:`str` of length > 1) or by code (passing a :obj:`str` of length 1).

        Passing :obj:`None` (default) to the `multicodecs` (resp. `multibases`) keyword argument means that all multicodecs (resp. multibases) will be loaded.
    """
    # pylint: disable = global-statement
    global _enabled_multicodecs
    global _enabled_multibases
    _multicodecs = {_validate_multicodec_data(data) for data in codecs} if codecs is not None else None
    _multibases = {_validate_multibase_data(data) for data in bases} if bases is not None else None
    if _multicodecs is not None:
        _multicodecs.update(_minimal_multicodecs)
    if _multibases is not None:
        _multibases.update(_minimal_multibases)
    _enabled_multicodecs = _multicodecs
    _enabled_multibases = _multibases

def _validate_multicodec_data(data: Union[int, str]) -> Union[int, str]:
    if isinstance(data, str):
        _validate_multicodec_name(data)
    elif isinstance(data, int):
        _validate_multicodec_code(data)
    else:
        raise TypeError(f"Expected multicodec name (str) or code (int), found {data} of type {type(data)}.")
    return data

def _validate_multicodec_name(name: str) -> None:
    if not re.match(r"^[a-z][a-z0-9_-]+$", name):
        raise ValueError(f"Invalid multicodec name {repr(name)}")

def _validate_multicodec_code(code: int) -> None:
    if code < 0:
        raise ValueError(f"Invalid multicodec code {repr(code)}.")

def _validate_multibase_data(data: str) -> str:
    if not isinstance(data, str):
        raise TypeError(f"Expected multibase name (str of len > 1) or code (str of len 1), found {data} of type {type(data)}.")
    if len(data) == 1:
        _validate_multibase_code(data)
    elif isinstance(data, int):
        _validate_multibase_name(data)
    return data

def _validate_multibase_name(name: str) -> None:
    if not re.match(r"^[a-z][a-z0-9_-]+$", name): # ensures len(name) > 1
        raise ValueError(f"Invalid multibase encoding name {repr(name)}")

def _validate_multibase_code(code: str) -> None:
    if len(code) != 1:
        raise ValueError("Multibase codes must be single-character strings or the hex digits '0x...' of a non-empty bytestring.")



r"""
    Utilities to load and build the multicodec table.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

import importlib.resources as importlib_resources
import json
from typing import Dict, Iterable, Set, Tuple, TYPE_CHECKING

from .config import _enabled_multicodecs

if TYPE_CHECKING:
    from multiformats.multicodec import Multicodec

def build_multicodec_tables(codecs: Iterable[Multicodec], *,
                            allow_private_use: bool = False) -> Tuple[Dict[int, Multicodec], Dict[str, Multicodec]]:
    """
        Creates code->multicodec and name->multicodec mappings from a finite iterable of multicodecs,
        returning the mappings.

        Example usage:

        >>> code_table, name_table = build_multicodec_tables(codecs)

        :param codecs: multicodecs to be registered
        :type codecs: iterable of :class:`Multicodec`
        :param allow_private_use: whether to allow multicodec entries with private use codes in ``range(0x300000, 0x400000)`` (default :obj:`False`)
        :type allow_private_use: :obj:`bool`, *optional*

        :raises ValueError: if ``allow_private_use`` and a multicodec with private use code is encountered
        :raises ValueError: if the same multicodec code is encountered multiple times, unless exactly one of the multicodecs
        has permanent status (in which case that codec is the one inserted in the table)
        :raises ValueError: if the same name is encountered multiple times

    """
    # pylint: disable = import-outside-toplevel
    from multiformats.multicodec.err import MulticodecValueError
    from multiformats.multicodec import Multicodec
    # validate(codecs, Iterable[Multicodec]) # TODO: not yet properly supported by typing-validation
    # validate(allow_private_use, bool)
    code_table: Dict[int, Multicodec] = {}
    name_table: Dict[str, Multicodec] = {}
    overwritten_draft_codes: Set[int] = set()
    for m in codecs:
        if not allow_private_use and m.is_private_use:
            raise MulticodecValueError(f"Private use multicodec not allowed: {m}")
        if m.code in code_table:
            if code_table[m.code].status == "permanent":
                if m.status == "draft":
                    # this draft code has been superseded by a permanent one, skip it
                    continue
                raise MulticodecValueError(f"Multicodec code {m.hexcode} appears multiple times in table.")
            if m.status != "permanent":
                # overwriting draft code with another draft code: dodgy, need to check at the end
                overwritten_draft_codes.add(m.code)
        code_table[m.code] = m
        if m.name in name_table:
            raise MulticodecValueError(f"Multicodec name {m.name} appears multiple times in table.")
        name_table[m.name] = m
    for code in overwritten_draft_codes:
        m = code_table[code]
        if m.status != "permanent":
            raise MulticodecValueError(f"Code {m.code} appears multiple times in table, "
                              "but none of the associated multicodecs is permanent.")
    return code_table, name_table

def load_multicodec_table() -> Tuple[Dict[int, Multicodec], Dict[str, Multicodec]]:
    """
        Returns code->encoding and name->encoding mappings created (via :func:`build_multicodec_tables`) from the local copy of `multicodec-table.json`.
        If a subset of multicodecs has been enabled, only those multicodecs are loaded.

        Example usage:

        >>> code_table, name_table = load_multicodec_table()

    """
    # pylint: disable = import-outside-toplevel
    from multiformats.multicodec import Multicodec
    with importlib_resources.open_text("multiformats_config", "multicodec-table.json", encoding="utf8") as _table_f:
        table_json = json.load(_table_f)
        multicodecs = (Multicodec(**row) for row in table_json)
        if _enabled_multicodecs is not None:
            multicodecs = (m for m in multicodecs if m.name in _enabled_multicodecs or m.code in _enabled_multicodecs)
        code_table, name_table = build_multicodec_tables(multicodecs)
    return code_table, name_table



"""
    Configuration for the `multiformats <https://github.com/hashberg-io/multiformats>`_ package.
"""

# Table information and configuration options for the 'multiformats' package.
# Copyright (C) 2022 Hashberg Ltd

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

from __future__ import annotations # See https://peps.python.org/pep-0563/

__version__ = "0.2.0"

from .config import enable

__all__ = ("enable", )



"""
   Pulls the current multicodec table from the [multicodec spec](https://github.com/multiformats/multicodec)
"""

import csv
import io
import json
import pprint
import textwrap

# not a dependency for the `multiformats` library
import requests

from multiformats import multicodec
from multiformats_config.multicodec import build_multicodec_tables

# Fetches and validates the new multicodec table from the multicodec spec GitHub repo:
multicodec_table_url = "https://github.com/multiformats/multicodec/raw/master/table.csv"
print("Fetching multicodec table from:")
print(multicodec_table_url)
print()

new_bytes = requests.get(multicodec_table_url).content
new_text = new_bytes.decode("utf-8")
print("Building new multicodec table...")
reader = csv.DictReader(io.StringIO(new_text))
multicodecs = (multicodec.Multicodec(**{k.strip(): v.strip() for k, v in _row.items()})
               for _row in reader)
new_table, _ = build_multicodec_tables(multicodecs)

# Loads and validates the current multicodec table:
print("Building current multicodec table...")
with open("multiformats_config/multicodec-table.csv", "r") as f:
    current_text = f.read()
reader = csv.DictReader(io.StringIO(current_text))
multicodecs = (multicodec.Multicodec(**{k.strip(): v.strip() for k, v in _row.items()})
               for _row in reader)
current_table, _ = build_multicodec_tables(multicodecs)

print()

# Displays added multicodecs, if any:
added = {
    code: m
    for code, m in new_table.items()
    if code not in current_table
}
if added:
    print(f"Added {len(added)} new multicodecs:")
    for m in sorted(added.values(), key=lambda m: m.code):
        print(textwrap.indent(pprint.pformat(m.to_json()), "  "))
else:
    print("Added no new multicodecs.")

# Displays removed multicodecs, if any:
removed = {
    code: m
    for code, m in current_table.items()
    if code not in new_table
}
if removed:
    print(f"Removed {len(added)} existing multicodecs:")
    for m in sorted(removed.values(), key=lambda m: m.code):
        print(textwrap.indent(pprint.pformat(m.to_json()), "  "))
else:
    print("Removed no existing multicodecs.")

# Displays changed multicodecs, if any:
changed = {
    code: (m, new_table[code])
    for code, m in current_table.items()
    if code in new_table and new_table[code] != m
}
if changed:
    print(f"Changed {len(added)} existing multicodecs:")
    for m_old, m_new in sorted(changed.values(), key=lambda pair: pair[0].code):
        print(f"  Changes in protocol {m_old.hexcode}:")
        for attr in ("name", "tag", "status", "description"):
            old_val, new_val = (getattr(m_old, attr), getattr(m_new, attr))
            if old_val != new_val:
                print(f"    {attr}: {repr(old_val)} -> {repr(new_val)}")
else:
    print("Changed no existing multicodecs.")

print()

# If the table has changed, prompts for update:
if added or removed or changed:
    answer = input("Would you like to update the multicodec table? (y/n) ")
    if answer.lower().startswith("y"):
        with open("multiformats_config/multicodec-table.csv", "w") as f:
            f.write(new_text)
        with open("multiformats_config/multicodec-table.json", "w") as f:
            table = [new_table[code].to_json() for code in sorted(new_table.keys())]
            json.dump(table, f, indent=4)
else:
    print("Nothing to update, exiting.")



"""
    Pulls the current multibase table from the [multibase spec](https://github.com/multiformats/multibase)
"""

import csv
import io
import json
import pprint
import textwrap

# not a dependency for the `multiformats` library
import requests

from multiformats import multibase
from multiformats_config.multibase import build_multibase_tables


# Fetches and validates the new multibase table from the multibase spec GitHub repo:
multibase_table_url = "https://github.com/multiformats/multibase/raw/master/multibase.csv"
print("Fetching multibase table from:")
print(multibase_table_url)
print()

new_bytes = requests.get(multibase_table_url).content
new_text = new_bytes.decode("utf-8")
print("Building new multibase table...")
reader = csv.DictReader(io.StringIO(new_text))
clean_rows = ({k.strip(): v.strip() for k, v in row.items()} for row in reader)
renamed_rows = ({(k if k != "encoding" else "name"): v for k, v in row.items()} for row in clean_rows)
encodings = (multibase.Multibase(**{k.strip(): v.strip() for k, v in _row.items()})
             for _row in renamed_rows)
new_table, _ = build_multibase_tables(encodings)

# Loads and validates the current multibase table:
print("Building current multibase table...")
with open("multiformats_config/multibase-table.csv", "r", encoding="utf8") as f:
    current_text = f.read()
reader = csv.DictReader(io.StringIO(current_text))
clean_rows = ({k.strip(): v.strip() for k, v in row.items()} for row in reader)
renamed_rows = ({(k if k != "encoding" else "name"): v for k, v in row.items()} for row in clean_rows)
encodings = (multibase.Multibase(**{k.strip(): v.strip() for k, v in _row.items()})
             for _row in renamed_rows)
current_table, _ = build_multibase_tables(encodings)

print()

# Displays added encodings, if any:
added = {
    code: m
    for code, m in new_table.items()
    if code not in current_table
}
if added:
    print(f"Added {len(added)} new encodings:")
    for m in sorted(added.values(), key=lambda m: m.code):
        print(textwrap.indent(pprint.pformat(m.to_json()), "  "))
else:
    print("Added no new encodings.")

# Displays removed encodings, if any:
removed = {
    code: m
    for code, m in current_table.items()
    if code not in new_table
}
if removed:
    print(f"Removed {len(added)} existing encodings:")
    for m in sorted(removed.values(), key=lambda m: m.code):
        print(textwrap.indent(pprint.pformat(m.to_json()), "  "))
else:
    print("Removed no existing encodings.")

# Displays changed encodings, if any:
changed = {
    code: (m, new_table[code])
    for code, m in current_table.items()
    if code in new_table and new_table[code] != m
}

if changed:
    print(f"Changed {len(added)} existing encodings:")
    for m_old, m_new in sorted(changed.values(), key=lambda pair: pair[0].code):
        print(f"  Changes in protocol {repr(m_old.code)}:")
        for attr in ("name", "status", "description"):
            old_val, new_val = (getattr(m_old, attr), getattr(m_new, attr))
            if old_val != new_val:
                print(f"    {attr}: {repr(old_val)} -> {repr(new_val)}")
else:
    print("Changed no existing encodings.")

print()

# If the table has changed, prompts for update:
if added or removed or changed:
    answer = input("Would you like to update the multibase table? (y/n) ")
    if answer.lower().startswith("y"):
        with open("multiformats_config/multibase-table.csv", "w", encoding="utf8") as f:
            f.write(new_text)
        with open("multiformats_config/multibase-table.json", "w", encoding="utf8") as f:
            table = [new_table[code].to_json() for code in sorted(new_table.keys())]
            json.dump(table, f, indent=4)
else:
    print("Nothing to update, exiting.")
